class PrivateKey(object):
    def __init__(self, data):
        """
        :param data:
        :type data: bytes
        """
        self._data = data  # type: bytes

    @property
    def data(self):
        return self._data

    def __eq__(self, other):
        return type(other) is PrivateKey and self.data == other.data



from dissononce.dh.x25519.x25519 import PublicKey
from ..proto import wa20_pb2
import axolotl_curve25519 as curve
import logging
import time

logger = logging.getLogger(__name__)


class CertMan(object):
    def __init__(self):
        self._pubkeys = {
            "WhatsAppLongTerm1": bytearray(
                [20, 35, 117, 87, 77, 10, 88, 113, 102, 170, 231, 30, 190, 81, 100, 55, 196, 162, 139,
                 115, 227, 105, 92, 108, 225, 247, 249, 84, 93, 168, 238, 107])
        }

    def is_valid(self, rs, certificate_data):
        """
        :param rs:
        :type rs: PublicKey
        :param certificate_data:
        :type certificate_data: bytes
        :return:
        :rtype:
        """
        cert = wa20_pb2.NoiseCertificate()
        cert.ParseFromString(certificate_data)
        cert_details = wa20_pb2.NoiseCertificate.Details()
        cert_details.ParseFromString(cert.details)

        logger.debug(
            "NoiseCertificate(signature=[%d bytes], serial=%d, issuer='%s', expires=%d, subject='%s', ""key=[%d bytes])"
            % (
                len(cert.signature), cert_details.serial, cert_details.issuer, cert_details.expires,
                cert_details.subject, len(cert_details.key)
            )
        )

        if cert_details.issuer not in self._pubkeys:
            logger.error("noise certificate issued by unknown source: issuer=%s" % cert_details.issuer)
            return False

        if curve.verifySignature(bytes(self._pubkeys[cert_details.issuer]), cert.details, cert.signature) != 0:
            logger.error("invalid signature on noise ceritificate; issuer=%s" % cert_details.issuer)
            return False

        if cert_details.key != rs.data:
            logger.error("noise certificate key does not match proposed server static key; issuer=%s" % cert_details.issuer)
            return False

        if cert_details.HasField("expires") and cert_details.expires < int(time.time()):
            logger.error("noise certificate expired; issuer=%s" % cert_details.issuer)
            return False

        return True



from .appversion import AppVersionConfig


class UserAgentConfig(object):
    PLATFORM_ANDROID = 0
    PLATFORM_IOS = 1
    PLATFORM_WINDOWS_PHONE = 2
    PLATFORM_PYTHON_CLIENT = 7

    STR_TEMPLATE = """UserAgentConfig(
        platform={platform},
        app_version={app_version},
        mcc={mcc},
        mnc={mnc},
        os_version={os_version},
        manufacturer={manufacturer},
        device={device},
        os_build_number={os_build_number},
        phone_id={phone_id},
        locale_lang={locale_lang},
        locale_country={locale_country}
    )"""

    def __init__(self,
                 platform,
                 app_version,
                 mcc, mnc,
                 os_version, manufacturer, device, os_build_number,
                 phone_id,
                 locale_lang,
                 locale_country
                 ):
        """
        :param platform:
        :type platform: int
        :param app_version:
        :type app_version: AppVersionConfig | str
        :param mcc:
        :type mcc: str
        :param mnc:
        :type mnc: str
        :param os_version:
        :type os_version: str
        :param manufacturer:
        :type manufacturer: str
        :param device:
        :type device: str
        :param os_build_number:
        :type os_build_number:
        :param phone_id:
        :type phone_id:
        :param locale_lang:
        :type locale_lang: str
        :param locale_country:
        :type locale_country: str
        """
        if type(app_version) is str:
            app_version = AppVersionConfig(app_version)
        self._platform = platform
        self._app_version = app_version
        self._mcc = mcc
        self._mnc = mnc
        self._os_version = os_version
        self._manufacturer = manufacturer
        self._device = device
        self._os_build_number = os_build_number
        self._phone_id = phone_id
        self._locale_lang = locale_lang
        self._locale_country = locale_country

    def __str__(self):
        return self.STR_TEMPLATE.format(
            platform=self.platform,
            app_version=self.app_version,
            mcc=self.mcc,
            mnc=self.mnc,
            os_version=self.os_version,
            manufacturer=self.manufacturer,
            device=self.device,
            os_build_number=self.os_build_number,
            phone_id=self.phone_id,
            locale_lang=self.locale_lang,
            locale_country=self.locale_country
        )

    @property
    def platform(self):
        return self._platform

    @property
    def app_version(self):
        return self._app_version
    @property
    def mcc(self):
        return self._mcc

    @property
    def mnc(self):
        return self._mnc

    @property
    def os_version(self):
        return self._os_version

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def device(self):
        return self._device

    @property
    def os_build_number(self):
        return self._os_build_number

    @property
    def phone_id(self):
        return self._phone_id

    @property
    def locale_lang(self):
        return self._locale_lang

    @property
    def locale_country(self):
        return self._locale_country



class ByteUtil(object):
    @staticmethod
    def split(input, first_len, second_len, third_len=None):
        """
        :param input:
        :type input: bytes
        :param first_len:
        :type first_len: int
        :param second_len:
        :type second_len: int
        :param third_len:
        :type third_len: int
        :return:
        :rtype:
        """
        parts = []
        parts.append(input[:first_len])
        parts.append(input[first_len:first_len + second_len])
        if third_len is not None:
            parts.append(input[first_len + second_len: first_len + second_len + third_len])
        return parts



from .useragent import UserAgentConfig


class ClientConfig(object):
    STR_TEMPLATE = """ClientConfig(
    username={username},
    passive={passive},
    useragent={useragent},
    pushname={pushname},
    short_connect={short_connect}
)"""

    def __init__(self,
                 username,
                 passive,
                 useragent,
                 pushname,
                 short_connect=True,
                 ):
        """
        :param username:
        :type username: int
        :param passive:
        :type passive: bool
        :param useragent:
        :type useragent: UserAgentConfig
        :param pushname:
        :type pushname: str
        :param short_connect:
        :type short_connect: bool
        """
        self._username = username
        self._passive = passive
        self._useragent = useragent
        self._pushname = pushname
        self._short_connect = short_connect

    def __str__(self):
        return self.STR_TEMPLATE.format(
            username=self.username,
            passive=self.passive,
            useragent=self.useragent,
            pushname=self.pushname,
            short_connect=self.short_connect
        )

    @property
    def username(self):
        return self._username

    @property
    def passive(self):
        return self._passive

    @property
    def useragent(self):
        return self._useragent

    @property
    def pushname(self):
        return self._pushname

    @property
    def short_connect(self):
        return self._short_connect



from .segmented import SegmentedStream

try:
    import Queue
except ImportError:
    import queue as Queue


class BlockingQueueSegmentedStream(SegmentedStream):

    EVENT_READ  = 1
    EVENT_WRITE = 2

    def __init__(self):
        self._readqueue = Queue.Queue()
        self._writequeue = Queue.Queue()
        self._events_callback = None

    def set_events_callback(self, events_callback):
        self._events_callback = events_callback

    def remove_events_callback(self):
        self._events_callback = None

    def put_read_segment(self, data):
        """
        :param data:
        :type data: bytes
        :return:
        :rtype:
        """
        self._readqueue.put(data)

    def get_write_segment(self):
        """
        :return:
        :rtype: bytes
        """
        return self._writequeue.get(block=True)

    def read_segment(self):
        if self._events_callback is not None:
            self._events_callback(self.EVENT_READ)

        return self._readqueue.get(block=True)

    def write_segment(self, data):
        self._writequeue.put(data)

        if self._events_callback is not None:
            self._events_callback(self.EVENT_WRITE)




import logging

__version__ = "0.1.2"
__author__ = "Tarek Galal"

logger = logging.getLogger(__name__)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(levelname).1s %(asctime)s %(name)s - %(message)s')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)



from .segmented import SegmentedStream


class DummySegmentedStream(SegmentedStream):
    def __init__(self, data):
        """
        :param data:
        :type data: list[bytes]
        """
        self._read = data  # type: list[bytes]
        self._sent = []  # type: list[bytes]

    def read_segment(self):
        """
        :return:
        :rtype: bytes
        """
        return self._read.pop(0)

    def write_segment(self, data):
        """
        :param data:
        :type data: bytes
        :return:
        :rtype:
        """
        self._sent.append(data)



from dissononce.processing.impl.symmetricstate import SymmetricState


class WASymmetricState(SymmetricState):
    """WA does not mix_hash if cipherstate does not yet have a key, therefore this custom WASymmetricState has to
    be used instead of default one."""
    def encrypt_and_hash(self, plaintext):
        ciphertext = self._cipherstate.encrypt_with_ad(self._h, plaintext)
        if self._cipherstate.has_key():
            self.mix_hash(ciphertext)
        return ciphertext



from socket import socket

from .arbitrary import ArbitraryStream


class SocketArbitraryStream(ArbitraryStream):
    def __init__(self, socket):
        """
        :param socket:
        :type socket: socket
        """
        self._socket = socket # type: socket

    def read(self, readsize):
        return self._socket.recv(readsize)

    def write(self, data):
        self._socket.send(data)



class AppVersionConfig(object):
    STR_TEMPLATE = """AppVersionConfig(
            primary={primary},
            secondary={secondary},
            tertiary={tertiary}
        )"""

    def __init__(self, version):
        """
        :param version:
        :type version: str
        """
        dissected = version.split('.')
        assert len(dissected) > 2, "version must be in format x.y.z"
        self._primary, self._secondary, self._tertiary = map(lambda v:int(v), dissected)

    def __str__(self):
        return self.STR_TEMPLATE.format(
            primary=self.primary,
            secondary=self.secondary,
            tertiary=self.tertiary
        )

    @property
    def primary(self):
        return self._primary

    @property
    def secondary(self):
        return self._secondary

    @property
    def tertiary(self):
        return self._tertiary



from .publickey import PublicKey
from .privatekey import PrivateKey

from dissononce.dh.x25519.x25519 import X25519DH
from dissononce.dh.x25519.keypair import KeyPair as X25519KeyPair


class KeyPair(object):
    def __init__(self, public, private):
        """
        :param public:
        :type public: PublicKey
        :param private:
        :type private: PrivateKey
        """
        self._public = public  # type: PublicKey
        self._private = private  # type: PrivateKey
        
    @property
    def public(self):
        return self._public

    @property
    def private(self):
        return self._private

    def __eq__(self, other):
        return type(other) is KeyPair and other.public == self.public and other.private == self.private

    @classmethod
    def generate(cls):
        """
        :return:
        :rtype: KeyPair
        """
        keypair = X25519DH().generate_keypair()
        return KeyPair(
            PublicKey(
                keypair.public.data
            ),
            PrivateKey(
                keypair.private.data
            )
        )

    @classmethod
    def from_bytes(cls, data):
        """
        :param data:
        :type data: bytes
        :return:
        :rtype: KeyPair
        """
        keypair = X25519KeyPair.from_bytes(data)
        return KeyPair(
            PublicKey(
                keypair.public.data
            ),
            PrivateKey(
                keypair.private.data
            )
        )




import struct

from .segmented import SegmentedStream
from ..arbitrary.arbitrary import ArbitraryStream


class WASegmentedStream(SegmentedStream):
    def __init__(self, dynamicstream):
        """
        :param dynamicstream:
        :type dynamicstream:  DataStream
        """
        self._datastream = dynamicstream # type: ArbitraryStream

    def read_segment(self):
        size = struct.unpack('>I', b"\x00" + self._datastream.read(3))[0] # type: int
        return self._datastream.read(size)

    def write_segment(self, data):
        if len(data) >= 16777216:
            raise ValueError("data too large to write; length=%d" % len(data))

        self._datastream.write(struct.pack('>I', len(data))[1:])
        self._datastream.write(data)



from dissononce.processing.cipherstate import CipherState
from .streams.segmented.segmented import SegmentedStream


class WANoiseTransport(object):
    def __init__(self, stream, send_cipherstate, recv_cipherstate):
        """
        :param stream:
        :type stream: SegmentedStream
        :param send_cipherstate:
        :type send_cipherstate: CipherState
        :param recv_cipherstate:
        :type recv_cipherstate: CipherState
        """
        self._stream = stream
        self._send_cipherstate = send_cipherstate
        self._recv_cipherstate = recv_cipherstate

    def send(self, plaintext):
        """
        :param plaintext:
        :type plaintext: bytes
        :return:
        :rtype:
        """
        ciphertext = self._send_cipherstate.encrypt_with_ad(b'', plaintext)
        self._stream.write_segment(ciphertext)

    def recv(self):
        """
        :return:
        :rtype: bytes
        """
        ciphertext = self._stream.read_segment()
        plaintext = self._recv_cipherstate.decrypt_with_ad(b'', ciphertext)
        return bytearray(plaintext)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wa20.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='wa20.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\nwa20.proto\"\x97\x11\n\rClientPayload\x12\x10\n\x08username\x18\x01 \x01(\x04\x12\x0f\n\x07passive\x18\x03 \x01(\x08\x12\x35\n\x0f\x63lient_features\x18\x04 \x03(\x0e\x32\x1c.ClientPayload.ClientFeature\x12,\n\nuser_agent\x18\x05 \x01(\x0b\x32\x18.ClientPayload.UserAgent\x12(\n\x08web_info\x18\x06 \x01(\x0b\x32\x16.ClientPayload.WebInfo\x12\x11\n\tpush_name\x18\x07 \x01(\t\x12\x12\n\nsession_id\x18\t \x01(\x0f\x12\x15\n\rshort_connect\x18\n \x01(\x08\x12\x30\n\x0c\x63onnect_type\x18\x0c \x01(\x0e\x32\x1a.ClientPayload.ConnectType\x12\x34\n\x0e\x63onnect_reason\x18\r \x01(\x0e\x32\x1c.ClientPayload.ConnectReason\x12\x0e\n\x06shards\x18\x0e \x03(\x0f\x12,\n\ndns_source\x18\x0f \x01(\x0b\x32\x18.ClientPayload.DnsSource\x12\x1d\n\x15\x63onnect_attempt_count\x18\x10 \x01(\x0f\x12\x39\n\x11ios_app_extension\x18\x1e \x01(\x0e\x32\x1e.ClientPayload.IosAppExtension\x1a\x9a\x01\n\tDnsSource\x12\x35\n\tdnsMethod\x18\x0f \x01(\x0e\x32\".ClientPayload.DnsSource.DnsMethod\x12\x11\n\tappCached\x18\x10 \x01(\x08\"C\n\tDnsMethod\x12\x0c\n\x08METHOD_0\x10\x00\x12\x0c\n\x08METHOD_1\x10\x01\x12\x0c\n\x08METHOD_2\x10\x02\x12\x0c\n\x08METHOD_3\x10\x03\x1a\xeb\x05\n\tUserAgent\x12<\n\x08platform\x18\x01 \x01(\x0e\x32!.ClientPayload.UserAgent.Platform:\x07\x41NDROID\x12\x38\n\x0b\x61pp_version\x18\x02 \x01(\x0b\x32#.ClientPayload.UserAgent.AppVersion\x12\x0b\n\x03mcc\x18\x03 \x01(\t\x12\x0b\n\x03mnc\x18\x04 \x01(\t\x12\x12\n\nos_version\x18\x05 \x01(\t\x12\x14\n\x0cmanufacturer\x18\x06 \x01(\t\x12\x0e\n\x06\x64\x65vice\x18\x07 \x01(\t\x12\x17\n\x0fos_build_number\x18\x08 \x01(\t\x12\x10\n\x08phone_id\x18\t \x01(\t\x12I\n\x0frelease_channel\x18\n \x01(\x0e\x32\'.ClientPayload.UserAgent.ReleaseChannel:\x07RELEASE\x12!\n\x19locale_language_iso_639_1\x18\x0b \x01(\t\x12)\n!locale_country_iso_3166_1_alpha_2\x18\x0c \x01(\t\x1aV\n\nAppVersion\x12\x0f\n\x07primary\x18\x01 \x01(\r\x12\x11\n\tsecondary\x18\x02 \x01(\r\x12\x10\n\x08tertiary\x18\x03 \x01(\r\x12\x12\n\nquaternary\x18\x04 \x01(\r\"\xb6\x01\n\x08Platform\x12\x0b\n\x07\x41NDROID\x10\x00\x12\x07\n\x03IOS\x10\x01\x12\x11\n\rWINDOWS_PHONE\x10\x02\x12\x0e\n\nBLACKBERRY\x10\x03\x12\x0f\n\x0b\x42LACKBERRYX\x10\x04\x12\x07\n\x03S40\x10\x05\x12\x07\n\x03S60\x10\x06\x12\x11\n\rPYTHON_CLIENT\x10\x07\x12\t\n\x05TIZEN\x10\x08\x12\x0e\n\nENTERPRISE\x10\t\x12\x0f\n\x0bPLATFORM_10\x10\n\x12\x0f\n\x0bPLATFORM_11\x10\x0b\"=\n\x0eReleaseChannel\x12\x0b\n\x07RELEASE\x10\x00\x12\x08\n\x04\x42\x45TA\x10\x01\x12\t\n\x05\x41LPHA\x10\x02\x12\t\n\x05\x44\x45\x42UG\x10\x03\x1a\xa7\x03\n\x07WebInfo\x12\x11\n\tref_token\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x38\n\x0cwebd_payload\x18\x03 \x01(\x0b\x32\".ClientPayload.WebInfo.WebdPayload\x1a\xbd\x02\n\x0bWebdPayload\x12\x1f\n\x17uses_participant_in_key\x18\x01 \x01(\x08\x12!\n\x19supports_starred_messages\x18\x02 \x01(\x08\x12\"\n\x1asupports_document_messages\x18\x03 \x01(\x08\x12\x1d\n\x15supports_url_messages\x18\x04 \x01(\x08\x12\x1c\n\x14supports_media_retry\x18\x05 \x01(\x08\x12\x1a\n\x12supports_e2e_image\x18\x06 \x01(\x08\x12\x1a\n\x12supports_e2e_video\x18\x07 \x01(\x08\x12\x1a\n\x12supports_e2e_audio\x18\x08 \x01(\x08\x12\x1d\n\x15supports_e2e_document\x18\t \x01(\x08\x12\x16\n\x0e\x64ocument_types\x18\n \x01(\t\"\x19\n\rClientFeature\x12\x08\n\x04NONE\x10\x00\"\xac\x01\n\x0b\x43onnectType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04WIFI\x10\x01\x12\x08\n\x04\x45\x44GE\x10\x64\x12\x08\n\x04IDEN\x10\x65\x12\x08\n\x04UMTS\x10\x66\x12\x08\n\x04\x45VDO\x10g\x12\x08\n\x04GPRS\x10h\x12\t\n\x05HSDPA\x10i\x12\t\n\x05HSUPA\x10j\x12\x08\n\x04HSPA\x10k\x12\x08\n\x04\x43\x44MA\x10l\x12\x0b\n\x07ONExRTT\x10m\x12\t\n\x05\x45HRPD\x10n\x12\x07\n\x03LTE\x10o\x12\t\n\x05HSPAP\x10p\"c\n\rConnectReason\x12\x0c\n\x08REASON_0\x10\x00\x12\x0c\n\x08REASON_1\x10\x01\x12\x0c\n\x08REASON_2\x10\x02\x12\x0c\n\x08REASON_3\x10\x03\x12\x0c\n\x08REASON_4\x10\x04\x12\x0c\n\x08REASON_5\x10\x06\"2\n\x0fIosAppExtension\x12\t\n\x05\x45XT_0\x10\x00\x12\t\n\x05\x45XT_1\x10\x01\x12\t\n\x05\x45XT_2\x10\x02\"\xea\x02\n\x10HandshakeMessage\x12\x33\n\x0c\x63lient_hello\x18\x02 \x01(\x0b\x32\x1d.HandshakeMessage.ClientHello\x12\x33\n\x0cserver_hello\x18\x03 \x01(\x0b\x32\x1d.HandshakeMessage.ServerHello\x12\x35\n\rclient_finish\x18\x04 \x01(\x0b\x32\x1e.HandshakeMessage.ClientFinish\x1a\x41\n\x0b\x43lientHello\x12\x11\n\tephemeral\x18\x01 \x01(\x0c\x12\x0e\n\x06static\x18\x02 \x01(\x0c\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x1a\x41\n\x0bServerHello\x12\x11\n\tephemeral\x18\x01 \x01(\x0c\x12\x0e\n\x06static\x18\x02 \x01(\x0c\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x1a/\n\x0c\x43lientFinish\x12\x0e\n\x06static\x18\x01 \x01(\x0c\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"\x90\x01\n\x10NoiseCertificate\x12\x0f\n\x07\x64\x65tails\x18\x01 \x01(\x0c\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x1aX\n\x07\x44\x65tails\x12\x0e\n\x06serial\x18\x01 \x01(\r\x12\x0e\n\x06issuer\x18\x02 \x01(\t\x12\x0f\n\x07\x65xpires\x18\x03 \x01(\x04\x12\x0f\n\x07subject\x18\x04 \x01(\t\x12\x0b\n\x03key\x18\x05 \x01(\x0c')
)



_CLIENTPAYLOAD_DNSSOURCE_DNSMETHOD = _descriptor.EnumDescriptor(
  name='DnsMethod',
  full_name='ClientPayload.DnsSource.DnsMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='METHOD_0', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='METHOD_1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='METHOD_2', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='METHOD_3', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=616,
  serialized_end=683,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_DNSSOURCE_DNSMETHOD)

_CLIENTPAYLOAD_USERAGENT_PLATFORM = _descriptor.EnumDescriptor(
  name='Platform',
  full_name='ClientPayload.UserAgent.Platform',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ANDROID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IOS', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WINDOWS_PHONE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLACKBERRY', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BLACKBERRYX', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='S40', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='S60', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PYTHON_CLIENT', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TIZEN', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTERPRISE', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PLATFORM_10', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PLATFORM_11', index=11, number=11,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1188,
  serialized_end=1370,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_USERAGENT_PLATFORM)

_CLIENTPAYLOAD_USERAGENT_RELEASECHANNEL = _descriptor.EnumDescriptor(
  name='ReleaseChannel',
  full_name='ClientPayload.UserAgent.ReleaseChannel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RELEASE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BETA', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALPHA', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEBUG', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1372,
  serialized_end=1433,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_USERAGENT_RELEASECHANNEL)

_CLIENTPAYLOAD_CLIENTFEATURE = _descriptor.EnumDescriptor(
  name='ClientFeature',
  full_name='ClientPayload.ClientFeature',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1861,
  serialized_end=1886,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_CLIENTFEATURE)

_CLIENTPAYLOAD_CONNECTTYPE = _descriptor.EnumDescriptor(
  name='ConnectType',
  full_name='ClientPayload.ConnectType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WIFI', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EDGE', index=2, number=100,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IDEN', index=3, number=101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UMTS', index=4, number=102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EVDO', index=5, number=103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GPRS', index=6, number=104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HSDPA', index=7, number=105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HSUPA', index=8, number=106,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HSPA', index=9, number=107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDMA', index=10, number=108,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ONExRTT', index=11, number=109,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EHRPD', index=12, number=110,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LTE', index=13, number=111,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HSPAP', index=14, number=112,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1889,
  serialized_end=2061,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_CONNECTTYPE)

_CLIENTPAYLOAD_CONNECTREASON = _descriptor.EnumDescriptor(
  name='ConnectReason',
  full_name='ClientPayload.ConnectReason',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REASON_0', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REASON_1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REASON_2', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REASON_3', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REASON_4', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REASON_5', index=5, number=6,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2063,
  serialized_end=2162,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_CONNECTREASON)

_CLIENTPAYLOAD_IOSAPPEXTENSION = _descriptor.EnumDescriptor(
  name='IosAppExtension',
  full_name='ClientPayload.IosAppExtension',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXT_0', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXT_1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXT_2', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2164,
  serialized_end=2214,
)
_sym_db.RegisterEnumDescriptor(_CLIENTPAYLOAD_IOSAPPEXTENSION)


_CLIENTPAYLOAD_DNSSOURCE = _descriptor.Descriptor(
  name='DnsSource',
  full_name='ClientPayload.DnsSource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dnsMethod', full_name='ClientPayload.DnsSource.dnsMethod', index=0,
      number=15, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='appCached', full_name='ClientPayload.DnsSource.appCached', index=1,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CLIENTPAYLOAD_DNSSOURCE_DNSMETHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=529,
  serialized_end=683,
)

_CLIENTPAYLOAD_USERAGENT_APPVERSION = _descriptor.Descriptor(
  name='AppVersion',
  full_name='ClientPayload.UserAgent.AppVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='primary', full_name='ClientPayload.UserAgent.AppVersion.primary', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='secondary', full_name='ClientPayload.UserAgent.AppVersion.secondary', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tertiary', full_name='ClientPayload.UserAgent.AppVersion.tertiary', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quaternary', full_name='ClientPayload.UserAgent.AppVersion.quaternary', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1099,
  serialized_end=1185,
)

_CLIENTPAYLOAD_USERAGENT = _descriptor.Descriptor(
  name='UserAgent',
  full_name='ClientPayload.UserAgent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='platform', full_name='ClientPayload.UserAgent.platform', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='app_version', full_name='ClientPayload.UserAgent.app_version', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mcc', full_name='ClientPayload.UserAgent.mcc', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mnc', full_name='ClientPayload.UserAgent.mnc', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='os_version', full_name='ClientPayload.UserAgent.os_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='manufacturer', full_name='ClientPayload.UserAgent.manufacturer', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device', full_name='ClientPayload.UserAgent.device', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='os_build_number', full_name='ClientPayload.UserAgent.os_build_number', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='phone_id', full_name='ClientPayload.UserAgent.phone_id', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='release_channel', full_name='ClientPayload.UserAgent.release_channel', index=9,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='locale_language_iso_639_1', full_name='ClientPayload.UserAgent.locale_language_iso_639_1', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='locale_country_iso_3166_1_alpha_2', full_name='ClientPayload.UserAgent.locale_country_iso_3166_1_alpha_2', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CLIENTPAYLOAD_USERAGENT_APPVERSION, ],
  enum_types=[
    _CLIENTPAYLOAD_USERAGENT_PLATFORM,
    _CLIENTPAYLOAD_USERAGENT_RELEASECHANNEL,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=686,
  serialized_end=1433,
)

_CLIENTPAYLOAD_WEBINFO_WEBDPAYLOAD = _descriptor.Descriptor(
  name='WebdPayload',
  full_name='ClientPayload.WebInfo.WebdPayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uses_participant_in_key', full_name='ClientPayload.WebInfo.WebdPayload.uses_participant_in_key', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_starred_messages', full_name='ClientPayload.WebInfo.WebdPayload.supports_starred_messages', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_document_messages', full_name='ClientPayload.WebInfo.WebdPayload.supports_document_messages', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_url_messages', full_name='ClientPayload.WebInfo.WebdPayload.supports_url_messages', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_media_retry', full_name='ClientPayload.WebInfo.WebdPayload.supports_media_retry', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_e2e_image', full_name='ClientPayload.WebInfo.WebdPayload.supports_e2e_image', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_e2e_video', full_name='ClientPayload.WebInfo.WebdPayload.supports_e2e_video', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_e2e_audio', full_name='ClientPayload.WebInfo.WebdPayload.supports_e2e_audio', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supports_e2e_document', full_name='ClientPayload.WebInfo.WebdPayload.supports_e2e_document', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='document_types', full_name='ClientPayload.WebInfo.WebdPayload.document_types', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1542,
  serialized_end=1859,
)

_CLIENTPAYLOAD_WEBINFO = _descriptor.Descriptor(
  name='WebInfo',
  full_name='ClientPayload.WebInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ref_token', full_name='ClientPayload.WebInfo.ref_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='ClientPayload.WebInfo.version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='webd_payload', full_name='ClientPayload.WebInfo.webd_payload', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CLIENTPAYLOAD_WEBINFO_WEBDPAYLOAD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1436,
  serialized_end=1859,
)

_CLIENTPAYLOAD = _descriptor.Descriptor(
  name='ClientPayload',
  full_name='ClientPayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='username', full_name='ClientPayload.username', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='passive', full_name='ClientPayload.passive', index=1,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='client_features', full_name='ClientPayload.client_features', index=2,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_agent', full_name='ClientPayload.user_agent', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='web_info', full_name='ClientPayload.web_info', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='push_name', full_name='ClientPayload.push_name', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='session_id', full_name='ClientPayload.session_id', index=6,
      number=9, type=15, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='short_connect', full_name='ClientPayload.short_connect', index=7,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connect_type', full_name='ClientPayload.connect_type', index=8,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connect_reason', full_name='ClientPayload.connect_reason', index=9,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shards', full_name='ClientPayload.shards', index=10,
      number=14, type=15, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dns_source', full_name='ClientPayload.dns_source', index=11,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connect_attempt_count', full_name='ClientPayload.connect_attempt_count', index=12,
      number=16, type=15, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ios_app_extension', full_name='ClientPayload.ios_app_extension', index=13,
      number=30, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CLIENTPAYLOAD_DNSSOURCE, _CLIENTPAYLOAD_USERAGENT, _CLIENTPAYLOAD_WEBINFO, ],
  enum_types=[
    _CLIENTPAYLOAD_CLIENTFEATURE,
    _CLIENTPAYLOAD_CONNECTTYPE,
    _CLIENTPAYLOAD_CONNECTREASON,
    _CLIENTPAYLOAD_IOSAPPEXTENSION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15,
  serialized_end=2214,
)


_HANDSHAKEMESSAGE_CLIENTHELLO = _descriptor.Descriptor(
  name='ClientHello',
  full_name='HandshakeMessage.ClientHello',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ephemeral', full_name='HandshakeMessage.ClientHello.ephemeral', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='static', full_name='HandshakeMessage.ClientHello.static', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payload', full_name='HandshakeMessage.ClientHello.payload', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2398,
  serialized_end=2463,
)

_HANDSHAKEMESSAGE_SERVERHELLO = _descriptor.Descriptor(
  name='ServerHello',
  full_name='HandshakeMessage.ServerHello',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ephemeral', full_name='HandshakeMessage.ServerHello.ephemeral', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='static', full_name='HandshakeMessage.ServerHello.static', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payload', full_name='HandshakeMessage.ServerHello.payload', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2465,
  serialized_end=2530,
)

_HANDSHAKEMESSAGE_CLIENTFINISH = _descriptor.Descriptor(
  name='ClientFinish',
  full_name='HandshakeMessage.ClientFinish',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='static', full_name='HandshakeMessage.ClientFinish.static', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='payload', full_name='HandshakeMessage.ClientFinish.payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2532,
  serialized_end=2579,
)

_HANDSHAKEMESSAGE = _descriptor.Descriptor(
  name='HandshakeMessage',
  full_name='HandshakeMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_hello', full_name='HandshakeMessage.client_hello', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='server_hello', full_name='HandshakeMessage.server_hello', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='client_finish', full_name='HandshakeMessage.client_finish', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_HANDSHAKEMESSAGE_CLIENTHELLO, _HANDSHAKEMESSAGE_SERVERHELLO, _HANDSHAKEMESSAGE_CLIENTFINISH, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2217,
  serialized_end=2579,
)


_NOISECERTIFICATE_DETAILS = _descriptor.Descriptor(
  name='Details',
  full_name='NoiseCertificate.Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='serial', full_name='NoiseCertificate.Details.serial', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='issuer', full_name='NoiseCertificate.Details.issuer', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expires', full_name='NoiseCertificate.Details.expires', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='subject', full_name='NoiseCertificate.Details.subject', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='NoiseCertificate.Details.key', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2638,
  serialized_end=2726,
)

_NOISECERTIFICATE = _descriptor.Descriptor(
  name='NoiseCertificate',
  full_name='NoiseCertificate',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='details', full_name='NoiseCertificate.details', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signature', full_name='NoiseCertificate.signature', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_NOISECERTIFICATE_DETAILS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2582,
  serialized_end=2726,
)

_CLIENTPAYLOAD_DNSSOURCE.fields_by_name['dnsMethod'].enum_type = _CLIENTPAYLOAD_DNSSOURCE_DNSMETHOD
_CLIENTPAYLOAD_DNSSOURCE.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD_DNSSOURCE_DNSMETHOD.containing_type = _CLIENTPAYLOAD_DNSSOURCE
_CLIENTPAYLOAD_USERAGENT_APPVERSION.containing_type = _CLIENTPAYLOAD_USERAGENT
_CLIENTPAYLOAD_USERAGENT.fields_by_name['platform'].enum_type = _CLIENTPAYLOAD_USERAGENT_PLATFORM
_CLIENTPAYLOAD_USERAGENT.fields_by_name['app_version'].message_type = _CLIENTPAYLOAD_USERAGENT_APPVERSION
_CLIENTPAYLOAD_USERAGENT.fields_by_name['release_channel'].enum_type = _CLIENTPAYLOAD_USERAGENT_RELEASECHANNEL
_CLIENTPAYLOAD_USERAGENT.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD_USERAGENT_PLATFORM.containing_type = _CLIENTPAYLOAD_USERAGENT
_CLIENTPAYLOAD_USERAGENT_RELEASECHANNEL.containing_type = _CLIENTPAYLOAD_USERAGENT
_CLIENTPAYLOAD_WEBINFO_WEBDPAYLOAD.containing_type = _CLIENTPAYLOAD_WEBINFO
_CLIENTPAYLOAD_WEBINFO.fields_by_name['webd_payload'].message_type = _CLIENTPAYLOAD_WEBINFO_WEBDPAYLOAD
_CLIENTPAYLOAD_WEBINFO.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD.fields_by_name['client_features'].enum_type = _CLIENTPAYLOAD_CLIENTFEATURE
_CLIENTPAYLOAD.fields_by_name['user_agent'].message_type = _CLIENTPAYLOAD_USERAGENT
_CLIENTPAYLOAD.fields_by_name['web_info'].message_type = _CLIENTPAYLOAD_WEBINFO
_CLIENTPAYLOAD.fields_by_name['connect_type'].enum_type = _CLIENTPAYLOAD_CONNECTTYPE
_CLIENTPAYLOAD.fields_by_name['connect_reason'].enum_type = _CLIENTPAYLOAD_CONNECTREASON
_CLIENTPAYLOAD.fields_by_name['dns_source'].message_type = _CLIENTPAYLOAD_DNSSOURCE
_CLIENTPAYLOAD.fields_by_name['ios_app_extension'].enum_type = _CLIENTPAYLOAD_IOSAPPEXTENSION
_CLIENTPAYLOAD_CLIENTFEATURE.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD_CONNECTTYPE.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD_CONNECTREASON.containing_type = _CLIENTPAYLOAD
_CLIENTPAYLOAD_IOSAPPEXTENSION.containing_type = _CLIENTPAYLOAD
_HANDSHAKEMESSAGE_CLIENTHELLO.containing_type = _HANDSHAKEMESSAGE
_HANDSHAKEMESSAGE_SERVERHELLO.containing_type = _HANDSHAKEMESSAGE
_HANDSHAKEMESSAGE_CLIENTFINISH.containing_type = _HANDSHAKEMESSAGE
_HANDSHAKEMESSAGE.fields_by_name['client_hello'].message_type = _HANDSHAKEMESSAGE_CLIENTHELLO
_HANDSHAKEMESSAGE.fields_by_name['server_hello'].message_type = _HANDSHAKEMESSAGE_SERVERHELLO
_HANDSHAKEMESSAGE.fields_by_name['client_finish'].message_type = _HANDSHAKEMESSAGE_CLIENTFINISH
_NOISECERTIFICATE_DETAILS.containing_type = _NOISECERTIFICATE
DESCRIPTOR.message_types_by_name['ClientPayload'] = _CLIENTPAYLOAD
DESCRIPTOR.message_types_by_name['HandshakeMessage'] = _HANDSHAKEMESSAGE
DESCRIPTOR.message_types_by_name['NoiseCertificate'] = _NOISECERTIFICATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ClientPayload = _reflection.GeneratedProtocolMessageType('ClientPayload', (_message.Message,), dict(

  DnsSource = _reflection.GeneratedProtocolMessageType('DnsSource', (_message.Message,), dict(
    DESCRIPTOR = _CLIENTPAYLOAD_DNSSOURCE,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:ClientPayload.DnsSource)
    ))
  ,

  UserAgent = _reflection.GeneratedProtocolMessageType('UserAgent', (_message.Message,), dict(

    AppVersion = _reflection.GeneratedProtocolMessageType('AppVersion', (_message.Message,), dict(
      DESCRIPTOR = _CLIENTPAYLOAD_USERAGENT_APPVERSION,
      __module__ = 'wa20_pb2'
      # @@protoc_insertion_point(class_scope:ClientPayload.UserAgent.AppVersion)
      ))
    ,
    DESCRIPTOR = _CLIENTPAYLOAD_USERAGENT,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:ClientPayload.UserAgent)
    ))
  ,

  WebInfo = _reflection.GeneratedProtocolMessageType('WebInfo', (_message.Message,), dict(

    WebdPayload = _reflection.GeneratedProtocolMessageType('WebdPayload', (_message.Message,), dict(
      DESCRIPTOR = _CLIENTPAYLOAD_WEBINFO_WEBDPAYLOAD,
      __module__ = 'wa20_pb2'
      # @@protoc_insertion_point(class_scope:ClientPayload.WebInfo.WebdPayload)
      ))
    ,
    DESCRIPTOR = _CLIENTPAYLOAD_WEBINFO,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:ClientPayload.WebInfo)
    ))
  ,
  DESCRIPTOR = _CLIENTPAYLOAD,
  __module__ = 'wa20_pb2'
  # @@protoc_insertion_point(class_scope:ClientPayload)
  ))
_sym_db.RegisterMessage(ClientPayload)
_sym_db.RegisterMessage(ClientPayload.DnsSource)
_sym_db.RegisterMessage(ClientPayload.UserAgent)
_sym_db.RegisterMessage(ClientPayload.UserAgent.AppVersion)
_sym_db.RegisterMessage(ClientPayload.WebInfo)
_sym_db.RegisterMessage(ClientPayload.WebInfo.WebdPayload)

HandshakeMessage = _reflection.GeneratedProtocolMessageType('HandshakeMessage', (_message.Message,), dict(

  ClientHello = _reflection.GeneratedProtocolMessageType('ClientHello', (_message.Message,), dict(
    DESCRIPTOR = _HANDSHAKEMESSAGE_CLIENTHELLO,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:HandshakeMessage.ClientHello)
    ))
  ,

  ServerHello = _reflection.GeneratedProtocolMessageType('ServerHello', (_message.Message,), dict(
    DESCRIPTOR = _HANDSHAKEMESSAGE_SERVERHELLO,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:HandshakeMessage.ServerHello)
    ))
  ,

  ClientFinish = _reflection.GeneratedProtocolMessageType('ClientFinish', (_message.Message,), dict(
    DESCRIPTOR = _HANDSHAKEMESSAGE_CLIENTFINISH,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:HandshakeMessage.ClientFinish)
    ))
  ,
  DESCRIPTOR = _HANDSHAKEMESSAGE,
  __module__ = 'wa20_pb2'
  # @@protoc_insertion_point(class_scope:HandshakeMessage)
  ))
_sym_db.RegisterMessage(HandshakeMessage)
_sym_db.RegisterMessage(HandshakeMessage.ClientHello)
_sym_db.RegisterMessage(HandshakeMessage.ServerHello)
_sym_db.RegisterMessage(HandshakeMessage.ClientFinish)

NoiseCertificate = _reflection.GeneratedProtocolMessageType('NoiseCertificate', (_message.Message,), dict(

  Details = _reflection.GeneratedProtocolMessageType('Details', (_message.Message,), dict(
    DESCRIPTOR = _NOISECERTIFICATE_DETAILS,
    __module__ = 'wa20_pb2'
    # @@protoc_insertion_point(class_scope:NoiseCertificate.Details)
    ))
  ,
  DESCRIPTOR = _NOISECERTIFICATE,
  __module__ = 'wa20_pb2'
  # @@protoc_insertion_point(class_scope:NoiseCertificate)
  ))
_sym_db.RegisterMessage(NoiseCertificate)
_sym_db.RegisterMessage(NoiseCertificate.Details)


# @@protoc_insertion_point(module_scope)



from ..proto.wa20_pb2 import HandshakeMessage


class NewRemoteStaticException(Exception):
    def __init__(self, server_hello):
        """
        :param server_hello:
        :type server_hello: HandshakeMessage.ServerHello
        """
        self._server_hello = server_hello # type: HandshakeMessage.ServerHello

    @property
    def server_hello(self):
        return self._server_hello



# -*- coding: utf-8 -*-
import consonance
from setuptools import find_packages, setup

setup(
    name='consonance',
    version=consonance.__version__,
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=['dissononce>=0.34.3', 'python-axolotl-curve25519', 'transitions', 'protobuf>=3.6.0'],
    license='GPL-3+',
    author='Tarek Galal',
    author_email='tare2.galal@gmail.com',
    description="WhatsApp's handshake implementation using Noise Protocol",
    long_description="This library implements WhatsApp's handshake process which makes use of "
                     "Noise Pipes with Curve25519, AES-GCM, and SHA256 from Noise Protocol",
    platforms='any',
    classifiers=['Development Status :: 3 - Alpha',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.5',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.3',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX :: Linux',
                 'Topic :: Security :: Cryptography',
                 'Topic :: Software Development :: Libraries :: Python Modules']
)



import logging

logger = logging.getLogger(__file__)


class ArbitraryStream(object):
    def read(self, readsize):
        """
        :param readsize:
        :type readsize: int
        :return:
        :rtype: bytes
        """

    def write(self, data):
        """
        :param data:
        :type data: bytes
        :return:
        :rtype:
        """



import random

from dissononce.processing.impl.handshakestate import HandshakeState
from dissononce.extras.processing.handshakestate_guarded import GuardedHandshakeState
from dissononce.extras.processing.handshakestate_switchable import SwitchableHandshakeState
from dissononce.processing.handshakepatterns.handshakepattern import HandshakePattern
from dissononce.processing.handshakepatterns.interactive.IK import IKHandshakePattern
from dissononce.processing.handshakepatterns.interactive.XX import XXHandshakePattern
from dissononce.processing.modifiers.fallback import FallbackPatternModifier
from dissononce.processing.impl.cipherstate import CipherState
from dissononce.cipher.aesgcm import AESGCMCipher
from dissononce.hash.sha256 import SHA256Hash
from dissononce.dh.keypair import KeyPair
from dissononce.dh.x25519.public import PublicKey
from dissononce.dh.private import PrivateKey
from dissononce.dh.x25519.x25519 import X25519DH
from dissononce.extras.dh.dangerous.dh_nogen import NoGenDH

from .dissononce_extras.processing.symmetricstate_wa import WASymmetricState
from .proto import wa20_pb2
from .streams.segmented.segmented import SegmentedStream
from .certman.certman import CertMan
from .exceptions.new_rs_exception import NewRemoteStaticException
from .config.client import ClientConfig
from .structs.publickey import PublicKey
from .util.byte import ByteUtil

import logging

logger = logging.getLogger(__name__)


class WAHandshake(object):
    def __init__(self, version_major, version_minor):
        self._prologue = b"WA" + bytearray([version_major, version_minor])
        self._handshakestate = None  # type: HandshakeState | None

    def perform(self, client_config, stream, s, rs=None, e=None):
        """
        :param client_config:
        :type client_config:
        :param stream:
        :type stream:
        :param s:
        :type s: consonance.structs.keypair.KeyPair
        :param rs:
        :type rs: consonance.structs.publickey.PublicKey | None
        :type e: consonance.structs.keypair.KeyPair | None
        :return:
        :rtype:
        """
        logger.debug("perform(client_config=%s, stream=%s, s=%s, rs=%s, e=%s)" % (
            client_config, stream, s, rs, e
        ))
        dh = X25519DH()
        if e is not None:
            dh = NoGenDH(dh, PrivateKey(e.private.data))
        self._handshakestate = SwitchableHandshakeState(
            GuardedHandshakeState(
                HandshakeState(
                    WASymmetricState(
                        CipherState(
                            AESGCMCipher()
                        ),
                        SHA256Hash()
                    ),
                    dh
                )
            )
        )  # type: SwitchableHandshakeState
        dissononce_s = KeyPair(
            PublicKey(s.public.data),
            PrivateKey(s.private.data)
        )
        dissononce_rs = PublicKey(rs.data) if rs else None
        client_payload = self._create_full_payload(client_config)
        logger.debug("Create client_payload=%s" % client_payload)
        if rs is not None:
            try:
                cipherstatepair = self._start_handshake_ik(stream, client_payload, dissononce_s, dissononce_rs)
            except NewRemoteStaticException as ex:
               cipherstatepair = self._switch_handshake_xxfallback(stream, dissononce_s, client_payload, ex.server_hello)
        else:
            cipherstatepair = self._start_handshake_xx(stream, client_payload, dissononce_s)

        return cipherstatepair

    @property
    def rs(self):
        return PublicKey(self._handshakestate.rs.data) if self._handshakestate.rs else None

    def _start_handshake_xx(self, stream, client_payload, s):
        """
        :param stream:
        :type stream: SegmentedStream
        :param client_payload:
        :type client_payload:
        :param s:
        :type s: KeyPair
        :return:
        :rtype:
        """
        self._handshakestate.initialize(
            handshake_pattern=XXHandshakePattern(),
            initiator=True,
            prologue=self._prologue,
            s=s
        )
        ephemeral_public = bytearray()
        self._handshakestate.write_message(b'', ephemeral_public)
        handshakemessage = wa20_pb2.HandshakeMessage()
        client_hello = wa20_pb2.HandshakeMessage.ClientHello()

        client_hello.ephemeral = bytes(ephemeral_public)
        handshakemessage.client_hello.MergeFrom(client_hello)
        stream.write_segment(handshakemessage.SerializeToString())

        incoming_handshakemessage = wa20_pb2.HandshakeMessage()
        incoming_handshakemessage.ParseFromString(stream.read_segment())

        if not incoming_handshakemessage.HasField("server_hello"):
            raise ValueError("Handshake message does not contain server hello!")

        server_hello = incoming_handshakemessage.server_hello

        payload_buffer = bytearray()
        self._handshakestate.read_message(
            server_hello.ephemeral + server_hello.static + server_hello.payload, payload_buffer
        )
        certman = CertMan()

        if certman.is_valid(self._handshakestate.rs, bytes(payload_buffer)):
            logger.debug("cert is valid")
        else:
            logger.error("cert is not valid")

        message_buffer = bytearray()
        cipherpair = self._handshakestate.write_message(client_payload.SerializeToString(), message_buffer)

        static, payload = ByteUtil.split(bytes(message_buffer), 48, len(message_buffer) - 48)
        client_finish = wa20_pb2.HandshakeMessage.ClientFinish()
        client_finish.static = static
        client_finish.payload = payload
        outgoing_handshakemessage = wa20_pb2.HandshakeMessage()
        outgoing_handshakemessage.client_finish.MergeFrom(client_finish)
        stream.write_segment(outgoing_handshakemessage.SerializeToString())

        return cipherpair

    def _start_handshake_ik(self, stream, client_payload, s, rs):
        """
        :param stream:
        :type stream: SegmentedStream
        :param s:
        :type s: KeyPair
        :param rs:
        :type rs: PublicKey
        :return:
        :rtype:
        """
        self._handshakestate.initialize(
            handshake_pattern=IKHandshakePattern(),
            initiator=True,
            prologue=self._prologue,
            s=s,
            rs=rs
        )
        message_buffer = bytearray()
        self._handshakestate.write_message(client_payload.SerializeToString(), message_buffer)
        ephemeral_public, static_public, payload = ByteUtil.split(bytes(message_buffer), 32, 48, len(message_buffer) - 32 + 48)
        handshakemessage = wa20_pb2.HandshakeMessage()
        client_hello = wa20_pb2.HandshakeMessage.ClientHello()

        client_hello.ephemeral = ephemeral_public
        client_hello.static = static_public
        client_hello.payload = payload
        handshakemessage.client_hello.MergeFrom(client_hello)

        stream.write_segment(handshakemessage.SerializeToString())

        incoming_handshakemessage = wa20_pb2.HandshakeMessage()
        incoming_handshakemessage.ParseFromString(stream.read_segment())

        if not incoming_handshakemessage.HasField("server_hello"):
            raise ValueError("Handshake message does not contain server hello!")

        server_hello = incoming_handshakemessage.server_hello

        if server_hello.HasField("static"):
            raise NewRemoteStaticException(server_hello)

        payload_buffer = bytearray()
        return self._handshakestate.read_message(
            server_hello.ephemeral + server_hello.static + server_hello.payload, payload_buffer
        )

    def _switch_handshake_xxfallback(self, stream, s, client_payload, server_hello):
        """
        :param handshake_pattern:
        :type handshake_pattern: HandshakePattern
        :param stream:
        :type stream: SegmentedStream
        :param s:
        :type s: KeyPair
        :param e:
        :type e: KeyPair
        :param client_payload:
        :type client_payload:
        :param server_hello:
        :type server_hello:
        :return:
        :rtype: tuple(CipherState,CipherState)
        """
        self._handshakestate.switch(
                handshake_pattern=FallbackPatternModifier().modify(XXHandshakePattern()),
                initiator=True,
                prologue=self._prologue,
                s=s
            )
        payload_buffer = bytearray()
        self._handshakestate.read_message(server_hello.ephemeral + server_hello.static + server_hello.payload, payload_buffer)
        certman = CertMan()
        if certman.is_valid(self._handshakestate.rs, bytes(payload_buffer)):
            logger.debug("cert is valid")
        else:
            logger.error("cert is not valid")

        message_buffer = bytearray()

        cipherpair = self._handshakestate.write_message(client_payload.SerializeToString(), message_buffer)

        static, payload = ByteUtil.split(bytes(message_buffer), 48, len(message_buffer) - 48)
        client_finish = wa20_pb2.HandshakeMessage.ClientFinish()
        client_finish.static = static
        client_finish.payload = payload
        outgoing_handshakemessage = wa20_pb2.HandshakeMessage()
        outgoing_handshakemessage.client_finish.MergeFrom(client_finish)
        stream.write_segment(outgoing_handshakemessage.SerializeToString())

        return cipherpair

    def _create_full_payload(self, client_config):
        """
        :param client_config:
        :type client_config: ClientConfig
        :return:
        :rtype: wa20_pb2.ClientPayload
        """
        client_payload = wa20_pb2.ClientPayload()
        user_agent = wa20_pb2.ClientPayload.UserAgent()
        user_agent_app_version = wa20_pb2.ClientPayload.UserAgent.AppVersion()

        user_agent.platform = client_config.useragent.platform
        user_agent.mcc = client_config.useragent.mcc
        user_agent.mnc = client_config.useragent.mnc
        user_agent.os_version = client_config.useragent.os_version
        user_agent.manufacturer = client_config.useragent.manufacturer
        user_agent.device = client_config.useragent.device
        user_agent.os_build_number = client_config.useragent.os_build_number
        user_agent.phone_id = client_config.useragent.phone_id
        user_agent.locale_language_iso_639_1 = client_config.useragent.locale_lang
        user_agent.locale_country_iso_3166_1_alpha_2 = client_config.useragent.locale_country

        user_agent_app_version.primary = client_config.useragent.app_version.primary
        user_agent_app_version.secondary = client_config.useragent.app_version.secondary
        user_agent_app_version.tertiary = client_config.useragent.app_version.tertiary

        user_agent.app_version.MergeFrom(user_agent_app_version)

        client_payload.username = client_config.username
        client_payload.passive = client_config.passive
        client_payload.push_name = client_config.pushname

        max_int = (2**32) / 2

        client_payload.session_id = random.randint(-max_int, max_int-1)
        client_payload.short_connect = client_config.short_connect
        client_payload.connect_type = 1
        client_payload.user_agent.MergeFrom(user_agent)

        return client_payload



from ..useragent import UserAgentConfig


class SamsungS9PUserAgentConfig(UserAgentConfig):
    DEFAULT_LOCALE_LANG = 'en'
    DEFAULT_LOCALE_COUNTRY = 'US'
    DEFAULT_MCC = '000'
    DEFAULT_MNC = '000'
    OS_VERSION = "8.0.0"
    OS_BUILD_NUMBER = "star2ltexx-user 8.0.0 R16NW G965FXXU1ARCC release-keys"
    MANUFACTURER = "samsung"
    DEVICE = "star2lte"

    def __init__(self,
                 app_version,
                 phone_id,
                 mcc=None, mnc=None,
                 locale_lang=None,
                 locale_country=None):
        super(SamsungS9PUserAgentConfig, self).__init__(
            platform=UserAgentConfig.PLATFORM_ANDROID,
            app_version=app_version,
            mcc=mcc or self.DEFAULT_MCC, mnc=mnc or self.DEFAULT_MNC,
            os_version=self.OS_VERSION,
            manufacturer=self.MANUFACTURER,
            device=self.DEVICE,
            os_build_number=self.OS_BUILD_NUMBER,
            phone_id=phone_id,
            locale_lang=locale_lang or self.DEFAULT_LOCALE_LANG,
            locale_country=locale_country or self.DEFAULT_LOCALE_COUNTRY
        )



from .config.client import ClientConfig
from .handshake import WAHandshake
from .streams.segmented.segmented import SegmentedStream
from .transport import WANoiseTransport
from transitions import Machine


class WANoiseProtocol(object):
    STATE_INIT = 'init'
    STATE_HANDSHAKE = 'handshake'
    STATE_TRANSPORT = 'transport'
    STATE_ERROR = 'error'
    STATES = [
        STATE_INIT,
        STATE_HANDSHAKE,
        STATE_TRANSPORT,
        STATE_ERROR
    ]
    TRANSITIONS = [
        ['start', STATE_INIT, STATE_HANDSHAKE],
        ['finish', STATE_HANDSHAKE, STATE_TRANSPORT],
        ['fail', STATE_HANDSHAKE, STATE_ERROR],
        ['fail', STATE_TRANSPORT, STATE_ERROR],
        ['reset', STATE_INIT, '='],
        ['reset', STATE_TRANSPORT, STATE_INIT],
        ['reset', STATE_HANDSHAKE, STATE_INIT],
        ['reset', 'error', STATE_INIT],
        ['send', STATE_TRANSPORT, '='],
        ['receive', STATE_TRANSPORT, '='],
        ['start', STATE_ERROR, STATE_HANDSHAKE]
    ]

    def __init__(self, version_major, version_minor, protocol_state_callbacks=None):
        """
        :param version_major:
        :type version_major: int
        :param version_minor:
        :type version_minor: int
        :param datastream
        :type datastream SegmentedStream
        """
        self._version_major = version_major
        self._version_minor = version_minor
        self._protocol_state_callbacks = protocol_state_callbacks
        self._rs = None
        self._machine = Machine(
            states=self.STATES, transitions=self.TRANSITIONS, initial='init',
            after_state_change=self._trigger_state_callback
        )  # type: Machine

        self._transport = None # type: WANoiseTransport
        self._last_triggered_state = None

    def _trigger_state_callback(self):
        if self._protocol_state_callbacks is not None and self._last_triggered_state != self._machine.state:
            self._last_triggered_state = self._machine.state
            self._protocol_state_callbacks(self._machine.state)

    @property
    def state(self):
        return self._machine.state

    @property
    def rs(self):
        return self._rs

    def start(self, stream, client_config, s, rs=None):
        """
        :param stream
        :type stream: SegmentedStream
        :param client_config:
        :type client_config: ClientConfig
        :return:
        :rtype:
        :param s:
        :type s: consonance.structs.keypair.KeyPair
        :param rs:
        :type rs: consonance.structs.publickey.PublicKey
        """
        self._machine.start()
        handshake = WAHandshake(self._version_major, self._version_minor)
        result = handshake.perform(client_config, stream, s, rs)

        if result is not None:
            self._rs = handshake.rs
            self._transport = WANoiseTransport(stream, result[0], result[1])
            self._machine.finish()
            return True
        else:
            self._machine.fail()

        return False

    def reset(self):
        self._machine.reset()
        self._transport = None

    def send(self, data):
        """
        :param data:
        :type data: bytes
        :param datastream
        :type SegmentedDataStream
        :return:
        :rtype:
        """
        self._machine.send()
        self._transport.send(data)

    def receive(self):
        """
        :param datastream
        :type SegmentedDataStream
        :return:
        :rtype: bytes
        """
        self._machine.receive()
        return self._transport.recv()

